#pragma once

/// @file
/// @brief A header including pretty much all the headers of the library.
/// Including this will have a significant compile time cost.

#include "mtx/identifiers.hpp"

#include "mtx/events.hpp"
#include "mtx/events/aliases.hpp"
#include "mtx/events/avatar.hpp"
#include "mtx/events/canonical_alias.hpp"
#include "mtx/events/create.hpp"
#include "mtx/events/guest_access.hpp"
#include "mtx/events/history_visibility.hpp"
#include "mtx/events/join_rules.hpp"
#include "mtx/events/member.hpp"
#include "mtx/events/name.hpp"
#include "mtx/events/pinned_events.hpp"
#include "mtx/events/power_levels.hpp"
#include "mtx/events/redaction.hpp"
#include "mtx/events/tag.hpp"
#include "mtx/events/topic.hpp"
#include "mtx/events/voip.hpp"

#include "mtx/events/messages/audio.hpp"
#include "mtx/events/messages/elementeffect.hpp"
#include "mtx/events/messages/emote.hpp"
#include "mtx/events/messages/file.hpp"
#include "mtx/events/messages/image.hpp"
#include "mtx/events/messages/notice.hpp"
#include "mtx/events/messages/text.hpp"
#include "mtx/events/messages/unknown.hpp"
#include "mtx/events/messages/video.hpp"

#include "mtx/user_interactive.hpp"

#include "mtx/requests.hpp"
#include "mtx/responses.hpp"
